/*
 * accept()/accept4() wrappers
 *
 * It's tricky for gnulib to work with SOCK_CLOEXEC/SOCK_NONBLOCK
 * in a forward-compatible way, so we'll handle accept4()-compatibility
 * for ourselves.
 *
 * Copyright (C) 2012 Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */

/*
 * we always define this because accept4() may fail with ENOSYS
 * if it was compiled/built on a newer system and run on an older one
 */
MOG_NOINLINE static int
mog_accept_old(int sockfd, struct sockaddr *addr, socklen_t *addrlen)
{
	int fd = accept(sockfd, addr, addrlen);

	if (fd >= 0) {
		/*
		 * don't bother with the nicer gnulib functions, we know
		 * these flags we want aren't set so save a fcntl()
		 * syscall in a frequently-called function.
		 *
		 * If we don't have a real accept4() syscall, don't
		 * bother setting O_CLOEXEC here.  Systems without
		 * accept4() will end up calling mog_cloexec_from()
		 * anyways in a fork()-ed child
		 */
		CHECK(int, 0, fcntl(fd, F_SETFL, O_NONBLOCK|O_RDWR));
	}
	return fd;
}

#if defined(HAVE_ACCEPT4)
static int accept_fallback(int, struct sockaddr *, socklen_t *);
static int
mog_accept4(int sockfd, struct sockaddr *addr, socklen_t *addrlen)
{
	int fd = accept4(sockfd, addr, addrlen, SOCK_CLOEXEC|SOCK_NONBLOCK);

	return fd >= 0 ? fd : accept_fallback(sockfd, addr, addrlen);
}

static int (*mog_accept)(int, struct sockaddr *, socklen_t *) = mog_accept4;

MOG_NOINLINE static int
accept_fallback(int sockfd, struct sockaddr *addr, socklen_t *addrlen)
{
	if (errno != ENOSYS) return -1;

	mog_accept = mog_accept_old;
	syslog(LOG_WARNING, "accept4() failed, falling back");

	return mog_accept_old(sockfd, addr, addrlen);
}
#else  /* HAVE_ACCEPT4 */
static int (*mog_accept)(int, struct sockaddr *, socklen_t *) = mog_accept_old;
#endif /* HAVE_ACCEPT4 */
