/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#define FD_PAD_SIZE ((size_t)128)
verify(sizeof(struct mog_fd) <= FD_PAD_SIZE);
static int max_fd;
static size_t fd_heaps;
static const size_t FD_PER_HEAP = 256;
static unsigned char **fd_map;
static pthread_mutex_t fd_lock = PTHREAD_MUTEX_INITIALIZER;

static inline struct mog_fd *aref(size_t fd)
{
	unsigned char *base = fd_map[fd / FD_PER_HEAP];

	return (struct mog_fd *)(base + (fd % FD_PER_HEAP) * FD_PAD_SIZE);
}

static void fd_map_atexit(void)
{
	while (fd_heaps-- > 0)
		free(fd_map[fd_heaps]);
	free(fd_map);
}

static void *my_memalign(size_t size)
{
	void *ptr;
	int err = posix_memalign(&ptr, FD_PAD_SIZE, size);

	switch (err) {
	case 0: return ptr;
	case ENOMEM: mog_oom();
	}

	die("posix_memalign() failed: %s\n", strerror(err));
	return NULL;
}

static void fd_map_init(void)
{
	long open_max = sysconf(_SC_OPEN_MAX);
	size_t slots = open_max / FD_PER_HEAP + 1;
	size_t size = slots * sizeof(void *);

	assert(fd_map == NULL && "fd_map reinitialized?");
	fd_map = my_memalign(size);
	atexit(fd_map_atexit);
}

MOG_NOINLINE static struct mog_fd * grow_ref(size_t fd)
{
	assert(fd < INT_MAX && "fd too large");
	CHECK(int, 0, pthread_mutex_lock(&fd_lock));

	if (!fd_map) fd_map_init();
	while (fd >= (size_t)mog_sync_fetch(&max_fd)) {
		unsigned char *base = my_memalign(FD_PAD_SIZE * FD_PER_HEAP);
		struct mog_fd *tmp;
		size_t i;

		for (i = 0; i < FD_PER_HEAP; i++) {
			tmp = (struct mog_fd *)(base + (i * FD_PAD_SIZE));
			tmp->fd_type = MOG_FD_TYPE_UNUSED;
		}

		fd_map[fd_heaps++] = base;
		(void)mog_sync_add_and_fetch(&max_fd, FD_PER_HEAP);
	}

	CHECK(int, 0, pthread_mutex_unlock(&fd_lock));

	return aref(fd);
}

struct mog_fd *mog_fd_get(int fd)
{
	assert(fd >= 0 && "FD is negative");
	if (MOG_LIKELY(fd < mog_sync_fetch(&max_fd)))
		return aref((size_t)fd);

	return grow_ref(fd);
}

/* optional... */
void mog_fd_put(struct mog_fd *validate, int fd)
{
	struct mog_fd *tmp;
	assert(fd >= 0 && "FD is negative");
	assert(fd < mog_sync_fetch(&max_fd) && "FD too small");
	tmp = aref(fd);

	assert(tmp == validate && "tried to put incorrect mog_fd back in");
	validate->fd_type = MOG_FD_TYPE_UNUSED;
}

void mog_fdmap_stop_queues(void)
{
	int i;

	for (i = 0; i < FD_PER_HEAP; ++i) {
		struct mog_fd *mfd = mog_fd_get(i);

		if (mfd->fd_type != MOG_FD_TYPE_QUEUE)
			continue;
		mog_thrpool_quit(&mfd->as.queue.thrpool);
		mog_close(mfd->fd);
	}
}
